# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import functools
import inspect
import json
import os
import shutil
import tempfile
import unittest
import warnings
from collections import namedtuple
from contextlib import contextmanager
from logging import getLogger
from marv_cli import create_loglevels

create_loglevels()
warnings.simplefilter('always', DeprecationWarning)

from marv_node import run as marv_node_run
marv_node_run.RAISE_IF_UNFINISHED
marv_node_run.RAISE_IF_UNFINISHED = True

#from . import marv
from ..io import get_logger as marv_get_logger
from ..io import get_stream as marv_get_stream
from ..io import pull as marv_pull
from ..node import node as marv_node
from ..run import run_nodes  # export via testing
from ..setid import SetID


StubDataset = namedtuple('StubDataset', 'setid name collection files time_added timestamp')
StubFile = namedtuple('StubFile', 'idx path missing mtime size')


KEEP = os.environ.get('KEEP')
log = getLogger(__name__)


@contextmanager
def temporary_directory(keep=KEEP):
    """Create and cleanup temporary directory"""
    tmpdir = tempfile.mkdtemp()
    log.debug('created temporary directory %r', tmpdir)
    try:
        yield tmpdir
    finally:
        if not keep:
            shutil.rmtree(tmpdir)
            log.debug('cleaned up temporary directory %r', tmpdir)
        else:
            log.debug('keeping temporary directory %r by request', tmpdir)


def make_dataset(files=None, setid=None, name=None, collection=None,
                 time_added=None, timestamp=None):
    setid = SetID(42) if setid is None else setid
    name = 'NAME' if name is None else name
    collection = 'COL' if collection is None else collection
    files = [StubFile(i, x, False, 42, os.stat(x).st_size)
             for i, x in enumerate([] if files is None else files)]
    time_added = 0 if time_added is None else time_added
    timestamp = 0 if timestamp is None else timestamp
    return StubDataset(setid=setid, name=name, collection=collection,
                       files=files, time_added=time_added,
                       timestamp=timestamp)


def make_sink(node):
    msgs = []

    @marv_node()
    def testsink():
        log = yield marv_get_logger()
        stream = yield marv_get_stream(node)
        log.debug('recv %r', stream)

        while True:
            msg = yield marv_pull(stream)
            log.debug('recv %r', msg)
            if msg is None:
                return
            msgs.append(msg)

    testsink._key = '{}-testsink'.format(node.abbrev)
    testsink.stream = msgs
    return testsink


class Spy(object):
    def __init__(self):
        self.requests = []


def make_spy(node):
    spy = Spy()
    requests = spy.requests
    real_invoke = node.invoke
    @functools.wraps(real_invoke)
    def invoke(inputs=None):
        gen = real_invoke(inputs)
        req = gen.next()
        while True:
            requests.append(req)
            send = yield req
            req = gen.send(send)
    node.invoke = invoke
    return spy


class TestCase(unittest.TestCase):
    BAGS = None
    MARV_TESTING_RECORD = os.environ.get('MARV_TESTING_RECORD')
    bags = None
    cleanup = None
    maxDiff = None
    scanroot = None

    @classmethod
    def setUpClass(cls):
        scanroot = tempfile.mkdtemp()
        @classmethod
        def cleanup(cls):
            shutil.rmtree(scanroot)
        cls.cleanup_class = cleanup

    @classmethod
    def tearDownClass(cls):
        cls.cleanup_class()

    def assertNodeOutput(self, output, node):
        name = node.name
        testfile = inspect.getmodule(type(self)).__file__
        dirpath = os.path.join(os.path.dirname(testfile), 'output')
        if not os.path.isdir(dirpath):
            os.mkdir(dirpath)
        outfile = os.path.join(dirpath, name + '.json')
        out = [x.to_dict(which=True) for x in output]
        if self.MARV_TESTING_RECORD:
            with open(outfile, 'w') as f:
                json.dump(out, f, sort_keys=True, indent=2)
        else:
            with open(outfile) as f:
                self.assertEqual(json.loads(json.dumps(out)), json.load(f))



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import OrderedDict, defaultdict
from itertools import count
from logging import getLogger

from capnp.lib.capnp import KjException

from marv_pycapnp import Wrapper
from .io import Abort
from .io import NEXT, PAUSED, RESUME, THEEND, TheEnd
from .io import SetHeader, CreateStream, Fork, GetRequested, GetStream
from .io import GetLogger, Task, MsgRequest, MakeFile
from .io import Pull, PullAll, Push
from .mixins import GenWrapperMixin, LoggerMixin
from .node import Keyed
from .stream import Handle, Msg


class MakeFileNotSupported(Exception):
    """Only persistent streams can make files."""


class Driver(Keyed, GenWrapperMixin, LoggerMixin):
    """Drives a generator."""
    key = None
    started = False
    stopped = False
    stream_creation = True  # False after first output Msg() TODO

    @property
    def key(self):
        return self.stream.key

    @property
    def key_abbrev(self):
        return self.stream.key_abbrev

    @property
    def setid(self):
        return self.stream.setid

    @property
    def node(self):
        return self.stream.node

    @property
    def name(self):
        return self.stream.name

    def __init__(self, stream, inputs=None):
        self.stream = stream
        self.streams = OrderedDict([(stream.handle, stream)])
        self.inputs = inputs
        self._requested_streams = []
        self._gen = self._run()

    def start(self):
        req = self.next()
        assert req is None
        return NEXT

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self.key_abbrev)

    def _run(self):
        self.started = True

        gen = self.node.invoke(self.inputs)
        assert hasattr(gen, 'send'), gen

        yield  # Wait for start signal before returning anything notable
        yield self.stream

        request_counter = count()
        msg_request_counter = defaultdict(count)
        next_msg_index_counter = defaultdict(int)
        send = None
        finished = False
        while not finished:
            try:
                request = gen.send(send)
                self.logdebug('got from node %s', type(request))
                request_idx = request_counter.next()
            except StopIteration:
                finished = True
            except Abort:
                finished = True

            if finished:
                for stream in self.streams.values():
                    if len(stream.cache) == 0:
                        yield stream.handle.msg(stream.handle)
                    if len(stream.cache) == 1:
                        self.logdebug('finishing empty stream')
                    if not stream.ended:
                        yield stream.handle.msg(THEEND)
                break

            output = None

            # preprocess
            # TODO:
            # - yield out.msg() valid? or enfore
            # - yield marv.push(out.msg(foo)) or
            # - yield marv.push(foo, out)
            if isinstance(request, Push):
                output = request.output
            elif isinstance(request, (Msg, Wrapper)):
                output = request

            # process
            if output:
                if len(self.stream.cache) == 0:
                    yield self.stream.handle.msg(self.stream.handle)

                # With first output, stream creation is done
                if self.stream_creation:
                    self.stream_creation = False

                msg = output
                if not isinstance(msg, Msg):
                    msg = self.stream.handle.msg(msg)
                else:
                    assert msg.handle.node is self.node
                    assert msg.handle.setid == self.setid
                schema = self.node.schema
                # TODO: handles should not be published by all?
                # this got introduced for merging streams
                if isinstance(msg._data, Handle):
                    # TODO: check that stream we are publishing to is a Group
                    assert self.node.group, self.node
                if schema is not None and \
                   not isinstance(msg._data, (Wrapper, Handle, TheEnd)):
                    try:
                        msg._data = Wrapper.from_dict(schema, msg._data)
                    except KjException as e:
                        from pprint import pformat
                        self.logerror(
                            'Schema violation for %s with data:\n%s\nschema: %s',
                            schema.schema.node.displayName,
                            pformat(msg._data),
                            schema.schema.node.displayName)
                        raise e
                signal = yield msg
                assert signal in (NEXT, RESUME), signal
                continue

            elif isinstance(request, Pull):
                handle = request.handle
                # if request.skip is never used the two indices remain the same
                msg_req_idx = msg_request_counter[handle].next()
                next_msg_idx = next_msg_index_counter[handle]  # + request.skip
                next_msg_index_counter[handle] = next_msg_idx + 1
                msg = yield MsgRequest(handle, next_msg_idx, self)
                assert isinstance(msg, Msg), msg
                assert msg.idx == next_msg_idx
                # TODO: switch to DONE
                send = None if msg.data is THEEND else msg.data
                if request.enumerate:
                    send = (msg_req_idx, send)
                continue

            elif isinstance(request, PullAll):
                send = []
                for handle in request.handles:
                    msg_req_idx = msg_request_counter[handle].next()
                    next_msg_idx = next_msg_index_counter[handle]
                    next_msg_index_counter[handle] = next_msg_idx + 1
                    msg = yield MsgRequest(handle, next_msg_idx, self)
                    assert isinstance(msg, Msg), msg
                    assert msg.idx == next_msg_idx
                    # TODO: switch to DONE
                    send.append(None if msg.data is THEEND else msg.data)
                continue

            elif isinstance(request, SetHeader):
                # TODO: should this be explicitly allowed/required?
                # Handles for non-header streams would be created right away
                assert len(self.stream.cache) == 0
                self.stream.handle.header = request.header.copy()
                yield self.stream.handle.msg(self.stream.handle)
                continue

            elif isinstance(request, MakeFile):
                if len(self.stream.cache) == 0:
                    yield self.stream.handle.msg(self.stream.handle)
                stream = self.streams[request.handle or self.stream.handle]
                try:
                    make_file = stream.make_file
                except AttributeError:
                    raise MakeFileNotSupported(stream)
                send = make_file(request.name)
                continue

            elif isinstance(request, Fork):
                parent_handle = self.stream.handle
                parent = self.streams[parent_handle]
                stream = parent.create_stream(name=request.name,
                                              group=request.group)
                fork = type(self)(stream, inputs=request.inputs)
                if len(self.stream.cache) == 0:
                    yield self.stream.handle.msg(self.stream.handle)
                yield fork
                send = None   # TODO: What should we send back?
                continue

            elif isinstance(request, GetStream):
                handle = Handle(request.setid or self.setid,
                                request.node, request.name)
                msg = yield MsgRequest(handle, -1, self)
                send = msg.data
                assert send == handle, (send, handle)
                continue  # TODO: why is this not covered?

            elif isinstance(request, CreateStream):
                kw = request._asdict()
                parent_handle = kw.pop('parent', None) or self.stream.handle
                parent = self.streams[parent_handle]
                stream = parent.create_stream(**kw)
                assert stream.name != 'default'
                assert stream.handle not in self.streams, stream
                self.streams[stream.handle] = stream
                if len(parent.cache) == 0:
                    yield parent.handle.msg(parent.handle)
                if len(self.stream.cache) == 0:
                    yield self.stream.handle.msg(self.stream.handle)
                yield stream
                yield stream.handle.msg(stream.handle)
                send = stream.handle
                continue

            elif isinstance(request, GetRequested):
                assert self.stream.group, (self, request)
                signal = yield PAUSED  # increase chances for completeness
                assert signal is RESUME
                send = list(self._requested_streams)
                self.stream_creation = False
                continue

            elif isinstance(request, GetLogger):
                send = getLogger('marv.node.{}'.format(self.key_abbrev))
                continue

            else:
                raise RuntimeError('Unknown request number {}: {!r} from {!r}'
                                   .format(request_idx + 1, request, self.node))
        self.stopped = True

    def add_stream_request(self, *handles):
        assert self.stream_creation
        assert self.node.group == 'ondemand'
        for handle in sorted(handles):
            if handle not in self._requested_streams:
                self._requested_streams.append(handle)
                self.logdebug('REQFILED %s', handle.key_abbrev)

    def destroy(self):
        for stream in self.streams.values():
            stream.destroy()

Task.register(Driver)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import struct
from base64 import b32decode, b32encode
from random import getrandbits


def decode_setid(encoded):
    """Decode setid as uint128"""
    try:
        lo, hi = struct.unpack('<QQ', b32decode(encoded.upper() + '======'))
    except struct.error:
        raise ValueError('Cannot decode {!r}'.format(encoded))
    return (hi << 64) + lo


def encode_setid(uint128):
    """Encode uint128 setid as stripped b32encoded string"""
    hi, lo = divmod(uint128, 2**64)
    return b32encode(struct.pack('<QQ', lo, hi))[:-6].lower()


class SetID(long):
    def __new__(cls, value, hi=None):
        if hi is not None:
            value = (hi << 64) + value
        value = decode_setid(value) if isinstance(value, basestring) else value
        return super(SetID, cls).__new__(cls, value)

    @classmethod
    def random(cls):
        return cls(getrandbits(128))

    @property
    def hi(self):
        return long(self) // 2**64

    @property
    def lo(self):
        return long(self) % 2**64

    @property
    def lohi(self):
        return tuple(long(x) for x in reversed(divmod(self, 2**64)))

    @property
    def abbrev(self):
        return str(self)[:10]

    def __repr__(self):
        return "SetID('{}')".format(self)

    def __str__(self):
        return encode_setid(self)

    def __unicode__(self):
        return encode_setid(self)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
import os

import pytest
from pathlib2 import Path

import marv
import marv.app
from marv.config import ConfigError
from marv.scanner import DatasetInfo
from marv.site import Site
from marv_nodes import dataset as dataset_node


DATADIR = Path(__file__).parent / 'data'
RECORD = os.environ.get('MARV_TESTING_RECORD')
SETIDS = [
    'l2vnfhfoe3z7ad7vclkd64tsqy',
    'vdls3sgw5yanuat4uepmhjwcpm',
    'e2oxlhpedjwked2llnnzir4tii',
    'phjg4ncymwbrbl4yx35bciqazq'
]

MARV_CONF = """
[marv]
collections = foo

[collection foo]
scanner = marv.tests.test_persist_without_type:scanner
scanroots = ./scanroots/foo
nodes =
    marv_nodes:dataset
    marv_nodes:summary_keyval
    marv_nodes:meta_table
    marv.tests.test_persist_without_type:notype
listing_columns =
    name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
    size       | Size   | filesize | (sum (get "dataset.files[:].size"))
    status     | Status | icon[]   | (status )
    tags       | Tags   | pill[]   | (tags )
    notype     | NoType | datetime | (get "notype.value")
"""


def scanner(dirpath, dirnames, filenames):
    return [DatasetInfo(x, [x]) for x in filenames]


@marv.node()
@marv.input('dataset', default=dataset_node)
def notype(dataset):
    dataset = yield marv.pull(dataset)
    with open(dataset.files[0].path) as f:
        yield marv.push({'value': int(f.read())})


@pytest.fixture(scope='function')
def site(tmpdir):
    flag = (tmpdir / 'TEST_SITE')
    flag.write('')

    marv_conf = (tmpdir / 'marv.conf')
    marv_conf.write(MARV_CONF)

    # make scanroots
    for sitename in ('foo',):
        for idx, name in enumerate(['a', 'b']):
            name = '{}_{}'.format(sitename, name)
            path = tmpdir / 'scanroots' / sitename / name
            path.write(str(idx), ensure=True)

    yield Site(marv_conf.strpath)


@pytest.fixture(scope='function')
def app(site):
    app = marv.app.create_app(site, init=True)
    app.testing = True
    with app.app_context():
        app = app.test_client()

        def get_json(*args, **kw):
            resp = app.get(*args, **kw)
            return json.loads(resp.data)

        app.get_json = get_json
        yield app


def test_fail_notype(app, site):
    with pytest.raises(ConfigError) as einfo:
        site.scan()



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
import unittest

from . import Store
from marv_node.testing import temporary_directory
from marv_node.setid import SetID
from marv_nodes import dataset


SETID = SetID(42)


@unittest.skip
class TestCase(unittest.TestCase):
    def test(self):
        with temporary_directory() as tmpdir:
            scanroot = os.path.join(tmpdir, 'scanroot')
            os.mkdir(scanroot)
            file1 = os.path.join(scanroot, 'file1')
            file2 = os.path.join(scanroot, 'file2')
            with open(file1, 'w') as f:
                f.write('1')
            with open(file2, 'w') as f:
                f.write('10')

            storedir = os.path.join(tmpdir, 'store')
            os.mkdir(storedir)
            store = Store(storedir)
            store.add_dataset(SETID, 'testset', [file1, file2])

            reader = store[(dataset, SETID)]
            self.assertEqual(reader.key, 'dataset')
            self.assertEqual(repr(reader),
                             '<Reader fiaaaaaaaaaaaaaaaaaaaaaaaa/dataset-1>')



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import capnp

from marv_pycapnp import Wrapper

from .test_wrapper_capnp import TestStruct


def test():
    builder = TestStruct.new_message()
    reader = builder.as_reader()
    wrapper = Wrapper(reader, streamdir=None, setdir=None)
    assert repr(wrapper) == '<Wrapper marv_pycapnp.tests.test_wrapper_capnp:TestStruct>'

    builder.text = u'\u03a8'
    assert wrapper.text == u'\u03a8'
    assert isinstance(wrapper.text, unicode)

    builder.data = u'\u03a8'.encode('utf-8')
    assert wrapper.data == u'\u03a8'.encode('utf-8')
    assert isinstance(wrapper.data, str)

    builder.textList = [u'\u03a8']
    assert wrapper.text_list == [u'\u03a8']
    assert wrapper.text_list[:] == [u'\u03a8']
    assert list(wrapper.text_list) == [u'\u03a8']
    assert isinstance(wrapper.text_list[0], unicode)
    assert repr(wrapper.textList) == "[u'\\u03a8']"

    builder.dataList = [u'\u03a8'.encode('utf-8')]
    assert wrapper.data_list == [u'\u03a8'.encode('utf-8')]
    assert wrapper.data_list[:] == [u'\u03a8'.encode('utf-8')]
    assert list(wrapper.data_list) == [u'\u03a8'.encode('utf-8')]
    assert isinstance(wrapper.data_list[0], str)
    assert repr(wrapper.dataList) == "['\\xce\\xa8']"

    builder.textListInList = [[u'\u03a8'], [u'\u03a8']]
    builder.dataListInList = [[u'\u03a8'.encode('utf-8')], [u'\u03a8'.encode('utf-8')]]
    assert all(isinstance(x, unicode) for lst in wrapper.textListInList for x in lst)
    assert all(isinstance(x, str) for lst in wrapper.dataListInList for x in lst)

    nested = Wrapper.from_dict(
        schema=TestStruct,
        data={
            'text': u'\u03a8',
            'data': u'\u03a8'.encode('utf-8'),
            'textList': [u'\u03a8'],
            'dataList': [u'\u03a8'.encode('utf-8')],
            'textListInList': [[u'\u03a8'], [u'\u03a8']],
            'dataListInList': [[u'\u03a8'.encode('utf-8')], [u'\u03a8'.encode('utf-8')]],
        }
    )
    builder.nestedList = [nested._reader]
    assert isinstance(wrapper.nested_list[0].text, unicode)
    assert isinstance(wrapper.nested_list[0].data, str)
    assert isinstance(wrapper.nested_list[0].textList[0], unicode)
    assert isinstance(wrapper.nested_list[0].dataList[0], str)

    builder.unionData = u'\u03a8'.encode('utf-8')
    assert isinstance(wrapper.union_data, str)

    builder.unionText = u'\u03a8'
    assert isinstance(wrapper.union_text, unicode)

    builder.union.data = u'\u03a8'.encode('utf-8')
    assert isinstance(wrapper.union.data, str)

    builder.union.text = u'\u03a8'
    assert isinstance(wrapper.union.text, unicode)

    builder.group.text = u'\u03a8'
    assert isinstance(wrapper.group.text, unicode)

    builder.group.data = u'\u03a8'.encode('utf-8')
    assert isinstance(wrapper.group.data, str)

    builder.enum = 'foo'
    assert wrapper.enum == 'foo'

    builder.enum = 'bar'
    assert wrapper.enum == 'bar'

    dct = wrapper.to_dict(which=True)
    assert dct == {
        'data': '\xce\xa8',
        'dataList': ['\xce\xa8'],
        'dataListInList': [['\xce\xa8'], ['\xce\xa8']],
        'enum': 'bar',
        'group': {
            'data': '\xce\xa8',
            'text': u'\u03a8'
        },
        'nestedList': [
            {
                'data': '\xce\xa8',
                'dataList': ['\xce\xa8'],
                'dataListInList': [['\xce\xa8'], ['\xce\xa8']],
                'enum': 'foo',
                'group': {
                    'data': '',
                    'text': u'',
                },
                'nestedList': [],
                'text': u'\u03a8',
                'textList': [u'\u03a8'],
                'textListInList': [[u'\u03a8'], [u'\u03a8']],
                'union': {
                    'text': u'',
                    '_which': 'text',
                },
                'unionText': u'',
                '_which': 'unionText',
            }
        ],
        'text': u'\u03a8',
        'textList': [u'\u03a8'],
        'textListInList': [[u'\u03a8'], [u'\u03a8']],
        'union': {
            'text': u'\u03a8',
            '_which': 'text',
        },
        'unionText': u'\u03a8',
        '_which': 'unionText',
    }

    dct = wrapper.to_dict()
    roundtrip = Wrapper.from_dict(TestStruct, dct)
    assert dct == roundtrip.to_dict()



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest

from ..run import run_nodes
from ..testing import make_dataset, make_sink, marv


dataset = make_dataset()
SETID = dataset.setid


class TestCase(unittest.TestCase):
    def test_run(self):
        @marv.node()
        @marv.input('offset', default=0)
        def root(offset):
            yield marv.push(10+offset)
            yield marv.push(20+offset)

        @marv.node()
        @marv.input('stream_a', default=root)
        def square(stream_a):
            while True:
                a = yield marv.pull(stream_a)
                if a is None:
                    break
                yield marv.push(a**2)

        @marv.node()
        @marv.input('stream_a', default=root)
        @marv.input('stream_b', default=root.clone(offset=5))
        @marv.input('stream_c', default=square)
        def add(stream_a, stream_b, stream_c):
            while True:
                a = yield marv.pull(stream_a)
                b = yield marv.pull(stream_b)
                c = yield marv.pull(stream_c)
                if a is None:
                    break
                yield marv.push(a+b+c)

        sinks = [make_sink(x) for x in [root, square, add]]
        run_nodes(dataset, sinks, {})
        self.assertEqual([x.stream for x in sinks], [
            [10, 20],
            [100, 400],
            [125, 445],
        ])

    def test_run_combined(self):
        @marv.node()
        @marv.input('offset', default=0)
        def root(offset):
            yield marv.push(10+offset)
            yield marv.push(20+offset)

        @marv.node()
        @marv.input('stream_a', default=root)
        def square(stream_a):
            while True:
                a = yield marv.pull(stream_a)
                if a is None:
                    break
                yield marv.push(a**2)

        @marv.node()
        @marv.input('stream_a', default=root)
        @marv.input('stream_b', default=root.clone(offset=5))
        @marv.input('stream_c', default=square)
        def add(stream_a, stream_b, stream_c):
            while True:
                a, b, c = yield marv.pull_all(stream_a, stream_b, stream_c)
                if a is None:
                    break
                yield marv.push(a+b+c)

        sinks = [make_sink(x) for x in [root, square, add]]
        run_nodes(dataset, sinks, {})
        self.assertEqual([x.stream for x in sinks], [
            [10, 20],
            [100, 400],
            [125, 445],
        ])

    def test_substream_subscription(self):
        @marv.node(group='ondemand')
        def source():
            out = yield marv.create_stream('foo')
            yield out.msg(1)

        msgs = []
        @marv.node()
        @marv.input('handle', default=marv.select(source, 'foo'))
        def consumer(handle):
            msg = yield marv.pull(handle)
            msgs.append(msg)

        sinks = [make_sink(x) for x in [consumer]]
        run_nodes(dataset, sinks, {})
        self.assertEqual(msgs, [1])



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest

from ..testing import marv
from ..node import InputNameCollision, Node


class TestCase(unittest.TestCase):
    def test_node_duplicate_input(self):
        with self.assertRaises(InputNameCollision):
            @marv.node()
            @marv.input('a')
            @marv.input('a')
            def no_inputs():
                yield

    @unittest.skip  # Do we want this?
    def test_node_without_inputs(self):
        @marv.node()
        def no_inputs():
            return 1
        self.assertEqual(no_inputs(), 1)

    def test_double_declare_node(self):
        with self.assertRaises(TypeError):
            @marv.node()
            @marv.node()
            def foo():
                yield  #pragma: nocoverage

    def test_node_repr(self):
        @marv.node()
        def foo():
            yield
        foo()
        self.assertEqual(
            repr(foo),
            '<Node foo.fy4oo6zcym>'
        )

    def test_comparisons(self):
        @marv.node()
        def a():
            yield

        @marv.node()
        def b():
            yield

        self.assertIs(type(a), type(b))
        self.assertIs(type(a), Node)
        self.assertIs(a, a)
        self.assertIsNot(a, b)

        self.assertLess(a.key, b.key)
        self.assertLess(a, b)
        self.assertLessEqual(a, b)
        self.assertLessEqual(a, a)
        self.assertEqual(a, a)
        self.assertNotEqual(a, b)
        self.assertGreater(b, a)
        self.assertGreaterEqual(b, a)
        self.assertGreaterEqual(b, b)

        self.assertEqual(a, a.clone())



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import deque
from itertools import count
from numbers import Integral
from types import NoneType

from .mixins import Keyed, LoggerMixin, Task, Request
from .setid import SetID


CACHESIZE = 50


class RequestedMessageTooOld(Exception):
    """Indicate a message requested from a stream is not in memory anymore"""


class Handle(Keyed):
    @property
    def key(self):
        return (self.setid, self.node, self.name)

    @property
    def key_abbrev(self):
        return '.'.join([self.setid.abbrev, self.node.abbrev, self.name])

    def __init__(self, setid, node, name, group=None, header=None):
        from .node import Node
        assert isinstance(setid, SetID), setid
        assert isinstance(node, Node), node
        self.setid = setid
        self.node = node
        self.name = name
        self.header = header or {}
        self.group = group if group is not None else self.node.group
        self._counter = count(-1)  # -1 will be the stream's handle

    def msg(self, __msg=None, **kw):
        from .io import THEEND
        assert (__msg is not None) ^ bool(kw), (__msg, kw)
        data = kw if __msg is None else __msg
        if self.group:
            assert isinstance(data, Handle) or data is THEEND, (self, data)
        return Msg(self._counter.next(), self, data)

    def finish(self):
        from .io import THEEND
        return self.msg(THEEND)

    def create_stream(self, name, group=False, **header):
        assert self.group
        from .io import CreateStream
        return CreateStream(self, name, group, header)

    def make_file(self, name):
        from .io import MakeFile
        return MakeFile(self, name)

    def __getattr__(self, name):
        try:
            return self.header[name]
        except KeyError:
            raise AttributeError(name)

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self.key_abbrev)


class Msg(Keyed):
    @property
    def key(self):
        return (self._idx, self._handle)

    @property
    def idx(self):
        return self._idx

    @property
    def handle(self):
        return self._handle

    @property
    def data(self):
        return self._data

    def __init__(self, idx=None, handle=None, data=None):
        assert isinstance(idx, (NoneType, Integral)), idx
        assert isinstance(handle, Handle), handle
        self._idx = idx
        self._handle = handle
        self._data = data

    def __iter__(self):
        return iter([self._idx, self._handle, self._data])

    def __repr__(self):
        from .io import THEEND
        flags = ' HANDLE' if isinstance(self.data, Handle) else \
                ' THEEND' if self.data is THEEND else ''  # data={!r}'.format(self.data)
        return 'Msg({}, {!r}{})'.format(self._idx, self.handle, flags)

Task.register(Msg)  # TODO: maybe not
Request.register(Msg)


class Stream(Keyed, LoggerMixin):
    # TODO: limit cache size
    CACHESIZE = CACHESIZE
    cache = None
    ended = None
    handle = None

    @property
    def key(self):
        return self.handle.key

    @property
    def key_abbrev(self):
        return self.handle.key_abbrev

    @property
    def setid(self):
        return self.handle.setid

    @property
    def node(self):
        return self.handle.node

    @property
    def name(self):
        return self.handle.name

    @property
    def group(self):
        return self.handle.group

    def info(self):
        return [repr(msg) for msg in self.cache]

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self.key_abbrev)


Task.register(Stream)


class VolatileStream(Stream):
    def __init__(self, handle, parent=None):
        self.handle = handle
        self.parent = parent
        self.cache = deque((), self.CACHESIZE)

    def add_msg(self, msg):
        from .io import THEEND
        assert msg.handle == self.handle
        assert msg.idx is not None
        if self.group:
            assert isinstance(msg.data, Handle) or msg.data is THEEND, (self, msg)
        else:
            assert not isinstance(msg.data, Handle) or msg.idx == -1, (self, msg)

        expected_idx = self.cache[0].idx + 1 if self.cache else -1
        assert msg.idx == expected_idx, (msg.idx, expected_idx)
        self.cache.appendleft(msg)

        if msg.data is THEEND:
            self.ended = True

        self.logdebug('added %r', msg)

    def get_msg(self, req):
        assert req.handle == self.handle

        offset = self.cache[0].idx - req.idx if self.cache else -1
        if offset < 0:
            return None
        try:
            msg = self.cache[offset]
        except IndexError:
            raise RequestedMessageTooOld(req, offset)
        assert msg.data is not None
        self.logdebug('return %r', msg)
        return msg

    def create_stream(self, name, group, header=None):
        # TODO: check name collision?
        handle = Handle(self.setid, self.node, name, group=group, header=header)
        return type(self)(handle, parent=self)

    def destroy(self):
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division

"""Access control lists

Use access control lists (ACLs) to control who can perform which
actions. An ACL is a function that returns a dictionary mapping route
function names to list of groups being allowed to access the
route. The special groups ``__authenticated__`` and
``__unauthenticated__`` are automatically assigned within
authenticated, resp. unauthenticated sessions.

"""

def authenticated():
    """Require authentication

    - require authentication for everything
    - only admins may delete datasets
    """
    return {
        'collection': ['__authenticated__'],
        'comment': ['__authenticated__'],
        'compare': ['__authenticated__'],
        'delete': ['admin'],
        'detail': ['__authenticated__'],
        'file_list': ['__authenticated__'],
        'get_partial_url': ['__authenticated__'],
        'get_stream_url': ['__authenticated__'],
        'stream': ['__authenticated__'],
        'tag': ['__authenticated__'],
    }


def public():
    """Allow public access

    - anyone can read anything
    - authenticated users can comment, tag and compare
    - only admins may delete datasets
    """
    return {
        'collection': ['__unauthenticated__', '__authenticated__'],
        'comment': ['__authenticated__'],
        'compare': ['__authenticated__'],
        'delete': ['admin'],
        'detail': ['__unauthenticated__', '__authenticated__'],
        'file_list': ['__unauthenticated__', '__authenticated__'],
        'get_partial_url': ['__unauthenticated__', '__authenticated__'],
        'get_stream_url': ['__unauthenticated__', '__authenticated__'],
        'stream': ['__unauthenticated__'],
        'tag': ['__authenticated__'],
    }



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import fcntl
import json
import os
import shutil
from collections import Mapping, defaultdict
from pathlib2 import Path

from marv_node.mixins import LoggerMixin
from marv_pycapnp import Wrapper
from .streams import PersistentStream, ReadStream


class DirectoryAlreadyExists(Exception):
    """Temporary directory for next generation of stream already exists

    This might mean another node run is in progress or aborted without
    proper cleanup.
    """


class Store(Mapping, LoggerMixin):
    def __init__(self, path, nodes):
        self.path = path
        self.pending = {}
        self.nodes = nodes
        self.name_by_node = {v: k for k, v in nodes.items()}

    def has_setid(self, setid):
        return os.path.isdir(os.path.join(self.path, str(setid)))

    def __getitem__(self, handle):
        setdir = os.path.join(self.path, str(handle.setid))
        name = self.name_by_node.get(handle.node, handle.node.name)
        symlink = os.path.join(setdir, name)
        try:
            gendir = os.readlink(symlink)
        except OSError:
            raise KeyError(handle)
        streamdir = os.path.join(setdir, gendir)
        if not os.path.exists(streamdir):
            raise KeyError(handle)
        with open(os.path.join(streamdir, 'streams.json')) as f:
            streams = json.load(f)
        if handle.name != 'default':
            streams = streams['streams'][handle.name]
        return ReadStream(handle, streamdir, setdir, info=streams)

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def add_dataset(self, dataset, exists_okay=None):
        # In case of restoring a database the set directory may already exist.
        setdir = os.path.join(self.path, str(dataset.setid))
        try:
            os.mkdir(setdir)
        except OSError as e:
            if not e.errno == 17 or not exists_okay:
                raise

    def create_stream(self, handle):
        assert handle.name == 'default', handle
        setdir = os.path.join(self.path, str(handle.setid))
        assert os.path.exists(setdir), setdir
        name = self.name_by_node.get(handle.node, handle.node.name)
        symlink = os.path.join(setdir, name)
        try:
            curgen = int(os.readlink(symlink).rsplit('-')[-1])
        except OSError:
            curgen = 0
        next_name = '{}-{}'.format(name, curgen + 1)
        nextdir = os.path.join(setdir, next_name)
        newlink = os.path.join(setdir, '.{}'.format(name))
        tmpdir = os.path.join(setdir, '.' + next_name)
        try:
            os.mkdir(tmpdir)
        except OSError:
            pass
        tmpdir_fd = os.open(tmpdir, os.O_RDONLY)
        try:
            fcntl.flock(tmpdir_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            os.close(tmpdir_fd)
            self.logerror('directory exists %r', tmpdir)
            raise DirectoryAlreadyExists(tmpdir)
        for path in Path(tmpdir).absolute().iterdir():
            if path.is_dir():
                shutil.rmtree(str(path))
            else:
                os.remove(str(path))
        self.logdebug('created directory %r', tmpdir)

        def commit(stream):
            assert not stream.group or stream.done == stream.streams.viewkeys()
            self.lognoisy('committing %r', nextdir)
            streams = self._streaminfo(stream)
            path = os.path.join(tmpdir, 'streams.json')
            fd = os.open(path, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
            f = os.fdopen(fd, 'w')
            json.dump(streams, f, indent=2, sort_keys=True)
            f.close()
            os.rename(tmpdir, nextdir)
            os.utime(nextdir, None)
            os.symlink(next_name, newlink)
            os.rename(newlink, symlink)
            del self.pending[stream]
            fcntl.flock(tmpdir_fd, fcntl.LOCK_UN)
            os.close(tmpdir_fd)

        stream = PersistentStream(handle, tmpdir, setdir=setdir, commit=commit)
        self.pending[stream] = (tmpdir, tmpdir_fd)
        return stream

    def _streaminfo(self, stream):
        return {'name': stream.name,
                'header': stream.handle.header,
                'streams': {x.name: self._streaminfo(x) for x in (stream.streams or {}).values()}}

    def load(self, setdir, node=None, nodename=None, default=()):
        assert bool(node) != bool(nodename)
        assert nodename != 'dataset'
        assert node.name != 'dataset'
        # TODO: handle substream fun
        name = nodename or self.name_by_node.get(node, node.name)
        nodedir = os.path.join(setdir, name)
        try:
            with open(os.path.join(nodedir, 'default-stream')) as f:
                msgs = node.schema.read_multiple_packed(f)
                return [Wrapper(x, None, setdir) for x in msgs]
        except IOError:
            if default is not ():
                return default
            raise



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractproperty
from logging import getLogger


class Keyed(object):
    __metaclass__ = ABCMeta

    @abstractproperty
    def key(self):
        return None  #pragma: nocoverage

    def __cmp__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return cmp(self.key, other.key)

    def __hash__(self):
        return hash((type(self), self.key))

    def __repr__(self):
        return '<{} key={!r}>'.format(type(self).__name__, self.key)


class GenWrapperMixin(object):
    _gen = None

    @property
    def close(self):
        return self._gen.close

    @property
    def next(self):
        return self._gen.next

    @property
    def send(self):
        return self._gen.send

    @property
    def throw(self):
        return self._gen.throw


class LoggerMixin(object):
    @property
    def logdebug(self):
        return self.log.debug

    @property
    def lognoisy(self):
        return self.log.noisy

    @property
    def logverbose(self):
        return self.log.verbose

    @property
    def loginfo(self):
        return self.log.info

    @property
    def logwarn(self):
        return self.log.warn

    @property
    def logerror(self):
        return self.log.error

    @property
    def log(self):
        logkey = ('marv', type(self).__name__.lower())
        if hasattr(self, 'key'):
            logkey += ((str(self.setid.abbrev),
                        self.node.name,
                        self.node.specs_hash[:10]) +
                       tuple(unicode(x) for x in self.key[2:]))
        return getLogger('.'.join(logkey))


class Task(object):
    __metaclass__ = ABCMeta
    __slots__ = ()

    def __repr__(self):
        return type(self).__name__


class Request(object):
    __metaclass__ = ABCMeta
    __slots__ = ()

    def __repr__(self):
        return type(self).__name__



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import fcntl
import json
import os
import shutil
from itertools import count, groupby, product
from logging import getLogger
from uuid import uuid4

import sqlalchemy as sqla
from flask import current_app as app
from pkg_resources import resource_filename

from marv_node.run import run_nodes
from marv_store import Store
from . import utils
from .collection import esc
from .collection import Collections
from .config import Config
from .model import STATUS_MISSING, STATUS_OUTDATED
from .model import Comment, Dataset, File, Group, Tag, User, dataset_tag, db


DEFAULT_NODES = """
marv_nodes:dataset
marv_nodes:meta_table
marv_nodes:summary_keyval
"""

DEFAULT_FILTERS = """
name       | Name       | substring         | string   | (get "dataset.name")
setid      | Set Id     | startswith        | string   | (get "dataset.id")
size       | Size       | lt le eq ne ge gt | filesize | (sum (get "dataset.files[:].size"))
status     | Status     | any all           | subset   | (status )
tags       | Tags       | any all           | subset   | (tags )
comments   | Comments   | substring         | string   | (comments )
files      | File paths | substring_any     | string[] | (get "dataset.files[:].path")
time_added | Added      | lt le eq ne ge gt | datetime | (get "dataset.time_added")
"""

DEFAULT_LISTING_COLUMNS = """
name       | Name   | route    | (detail_route (get "dataset.id") (get "dataset.name"))
size       | Size   | filesize | (sum (get "dataset.files[:].size"))
status     | Status | icon[]   | (status )
tags       | Tags   | pill[]   | (tags )
time_added | Added  | datetime | (get "dataset.time_added")
"""

DEFAULT_LISTING_SUMMARY = """
datasets | datasets | int       | (len (rows ))
size     | size     | filesize  | (sum (rows "size" 0))
"""

DEFAULT_DETAIL_SUMMARY_WIDGETS = """
summary_keyval
meta_table
"""

log = getLogger(__name__)


class UnknownNode(Exception):
    pass


def make_config(siteconf):
    sitedir = os.path.dirname(siteconf)
    defaults = {
        'marv': {
            'acl': 'marv_webapi.acls:authenticated',
            'dburi': 'sqlite:///' + os.path.join(sitedir, 'db', 'db.sqlite'),
            'frontenddir': os.path.join(sitedir, 'frontend'),
            'oauth': '',
            'reverse_proxy': None,
            'sessionkey_file': os.path.join(sitedir, 'sessionkey'),
            'staticdir': os.path.join(resource_filename('marv_ludwig', 'static')),
            'storedir': os.path.join(sitedir, 'store'),
            'window_title': '',
        },
        'collection': {
            'compare': None,
            'detail_summary_widgets': DEFAULT_DETAIL_SUMMARY_WIDGETS,
            'detail_sections': '',
            'detail_title': '(get "dataset.name")',
            'filters': DEFAULT_FILTERS,
            'listing_columns': DEFAULT_LISTING_COLUMNS,
            'listing_sort': '| ascending',
            'listing_summary': DEFAULT_LISTING_SUMMARY,
            'nodes': DEFAULT_NODES,
        }
    }
    required = {
        'marv': [
            'collections',
        ],
        'collection': [
            'scanner',
            'scanroots',
        ],
    }
    schema = {
        'marv': {
            'acl': 'find_obj',
            'collections': 'space_separated_list',
            'oauth': 'lines',
            'staticdir': 'path',
            'storedir': 'path',
        },
        'collection': {
            'compare': 'find_obj',
            'detail_title': 'function',
            'detail_summary_widgets': 'lines',
            'detail_sections': 'lines',
            'filters': 'lines',
            'listing_columns': 'lines',
            'listing_sort': 'pipe_separated_list',
            'listing_summary': 'lines',
            'nodes': 'lines',
            'scanroots': 'path_lines',
        },
    }
    return Config.from_file(siteconf, defaults=defaults,
                            required=required, schema=schema)


class Site(object):
    def __init__(self, siteconf):
        self.config = make_config(siteconf)
        # TODO: maybe ordereddict for meta, or generate multiple keys
        self.config.marv.oauth = dict(
            (i.split(' ')[0], [s.strip() for s in i.split('|')]
        ) for i in self.config.marv.oauth)
        self.collections = Collections(config=self.config)

    def load_for_web(self):
        _ = [getattr(x, y) and None for x, y, in product(
            self.collections.values(),
            ('compare', 'filter_specs', 'listing_columns', 'model',
             'sortcolumn', 'sortorder', 'summary_items'))]

    def init(self):
        try:
            fd = os.open(self.config.marv.sessionkey_file,
                         os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
        except OSError as e:
            if e.errno != 17:
                raise
        else:
            with os.fdopen(fd, 'w') as f:
                f.write(str(uuid4()))
            log.verbose('Generated %s', self.config.marv.sessionkey_file)

        try:
            os.mkdir(self.config.marv.storedir)
            log.verbose('Created %s', self.config.marv.storedir)
        except OSError as e:
            if e.errno != 17:
                raise

        try:
            dbdir = os.path.dirname(self.config.marv.dburi.replace('sqlite:///', ''))
            os.mkdir(dbdir)
            log.verbose('Created %s', dbdir)
        except OSError as e:
            if e.errno != 17:
                raise

        # TODO: maybe this should be an explicit method...
        # load models
        _ = [x.model for x in self.collections.values()]

        prefixes = ['listing_{}'.format(col) for col in self.collections.keys()]
        result = db.session.execute('SELECT name FROM sqlite_master WHERE type="table";')
        tables = {name for name in [x[0] for x in result]
                  if any(name.startswith(prefix) for prefix in prefixes)}
        for table in [x for x in tables if x.endswith('_sec')]:
            db.session.execute('DROP TABLE {};'.format(table))
            tables.discard(table)
        for table in [x for x in tables if x not in prefixes]:
            db.session.execute('DROP TABLE {};'.format(table))
            tables.discard(table)
        for table in tables:
            db.session.execute('DROP TABLE {};'.format(table))

        db.create_all()
        for name in ('admin', ):
            if not Group.query.filter(Group.name==name).count():
                db.session.add(Group(name=name))
                db.session.commit()

        try:
            with open('users', 'r') as usersfile:
                users = json.load(usersfile)
                for name, password in users.items():
                    if not User.query.filter(User.name==name).count():
                        db.session.add(User(name=name, password=password))
                        db.session.commit()
            os.rename('users', 'users.imported')
        except IOError:
            pass

        log.verbose('Initialized database %s', app.config['SQLALCHEMY_DATABASE_URI'])
        for col in self.collections.keys():
            collection = self.collections[col]
            loop = count()
            batchsize = 50
            # TODO: increase verbosity and probably introduce --reinit
            while True:
                batch = db.session.query(Dataset)\
                                  .filter(Dataset.collection == col)\
                                  .options(db.joinedload(Dataset.files))\
                                  .limit(batchsize)\
                                  .offset(batchsize*loop.next())\
                                  .all()
                for dataset in batch:
                    collection.render_detail(dataset)
                if batch:
                    collection.update_listings(batch)
                if len(batch) < batchsize:
                    break
        log.info('Initialized from %s', self.config.filename)

    def cleanup_discarded(self):
        collections = self.collections
        query = db.session.query(Dataset.collection, Dataset.id)\
                          .filter(Dataset.discarded.is_(True))\
                          .order_by(Dataset.collection)
        datasets = [(col, [x[1] for x in ids])
                    for col, ids in groupby(query, lambda x: x[0])]
        if not datasets:
            return

        for col, ids in datasets:
            collection = collections[col]
            model = collection.model
            listing = model.Listing.__table__
            for secondary in model.secondaries.values():
                stmt = secondary.delete()\
                                .where(secondary.c.listing_id.in_(ids))
                db.session.execute(stmt)

            stmt = listing.delete()\
                          .where(listing.c.id.in_(ids))
            db.session.execute(stmt)

            stmt = dataset_tag.delete()\
                              .where(dataset_tag.c.dataset_id.in_(ids))
            db.session.execute(stmt)

            db.session.query(Comment)\
                      .filter(Comment.dataset_id.in_(ids))\
                      .delete(synchronize_session=False)

            db.session.query(File)\
                      .filter(File.dataset_id.in_(ids))\
                      .delete(synchronize_session=False)

            db.session.query(Dataset)\
                      .filter(Dataset.id.in_(ids))\
                      .delete(synchronize_session=False)

        db.session.commit()
        # TODO: Cleanup corresponding store paths

    def cleanup_relations(self):
        """Cleanup listing relations"""
        collections = self.collections
        for relation in [x for col in collections.values()
                         for x in col.model.relations.values()]:
            db.session.query(relation)\
                       .filter(~relation.listing.any())\
                       .delete(synchronize_session=False)
        db.session.commit()

    def cleanup_tags(self):
        db.session.query(Tag)\
                  .filter(~Tag.datasets.any())\
                  .delete(synchronize_session=False)
        db.session.commit()

    def restore_database(self, datasets=None, users=None):
        for user in users or []:
            user.setdefault('realm', 'marv')
            user.setdefault('realmuid', '')
            groups = user.pop('groups', [])
            app.um.user_add(_restore=True, **user)
            for grp in groups:
                try:
                    app.um.group_adduser(grp, user['name'])
                except ValueError:
                    app.um.group_add(grp)
                    app.um.group_adduser(grp, user['name'])

        for key, sets in (datasets or {}).items():
            key = self.collections.keys()[0] if key == 'DEFAULT_COLLECTION' else key
            self.collections[key].restore_datasets(sets)

    def listtags(self, collections=None):
        query = db.session.query(Tag.value)
        if collections:
            query = query.filter(Tag.collection.in_(collections))
        query = query.group_by(Tag.value)\
                     .order_by(Tag.value)
        tags = [x[0] for x in query]
        return tags

    def query(self, collections=None, discarded=None, outdated=None,
              path=None, tags=None, abbrev=None, missing=None):
        abbrev = 10 if abbrev is True else abbrev
        discarded = bool(discarded)
        query = db.session.query(Dataset.setid)

        if collections:
            query = query.filter(Dataset.collection.in_(collections))

        if outdated:
            query = query.filter(Dataset.status.op('&')(STATUS_OUTDATED) == STATUS_OUTDATED)

        query = query.filter(Dataset.discarded.is_(discarded))

        if path:
            relquery = db.session.query(File.dataset_id)\
                                 .filter(File.path.like('{}%'.format(esc(path)), escape='$'))\
                                 .group_by(File.dataset_id)
            query = query.filter(Dataset.id.in_(relquery.subquery()))

        if missing:
            relquery = db.session.query(File.dataset_id)\
                                 .filter(File.missing.is_(True))\
                                 .group_by(File.dataset_id)
            query = query.filter(Dataset.id.in_(relquery.subquery()))

        if tags:
            relquery = db.session.query(dataset_tag.c.dataset_id)\
                                 .join(Tag)\
                                 .filter(Tag.value.in_(tags))
            query = query.filter(Dataset.id.in_(relquery.subquery()))

        setids = [x[0][:abbrev] if abbrev else x[0]
                  for x in query.order_by(Dataset.setid)]
        return setids

    def run(self, setid, selected_nodes=None, deps=None, force=None, keep=None,
            force_dependent=None, update_detail=None, update_listing=None,
            excluded_nodes=None, cachesize=None):
        assert not force_dependent or selected_nodes

        excluded_nodes = set(excluded_nodes or [])
        dataset = Dataset.query.filter(Dataset.setid == str(setid))\
                               .options(db.joinedload(Dataset.files))\
                               .one()
        collection = self.collections[dataset.collection]
        selected_nodes = set(selected_nodes or [])
        if not (selected_nodes or update_listing or update_detail):
            selected_nodes.update(collection.listing_deps)
            selected_nodes.update(collection.detail_deps)
        persistent = collection.nodes
        try:
            nodes = {persistent[name] if not ':' in name else utils.find_obj(name)
                     for name in selected_nodes
                     if name not in excluded_nodes
                     if name != 'dataset'}
        except KeyError as e:
            raise UnknownNode(dataset.collection, e.args[0])

        if force_dependent:
            nodes.update(x for name in selected_nodes
                         for x in persistent[name].dependent)
        nodes = sorted(nodes)

        storedir = app.site.config.marv.storedir
        store = Store(storedir, persistent)

        changed = False
        try:
            if nodes:
                changed = run_nodes(dataset, nodes, store, force=force,
                                    persistent=persistent,
                                    deps=deps, cachesize=cachesize)
        except:
            raise
        else:
            if changed or update_detail:
                collection.render_detail(dataset)
                log.verbose('%s detail rendered', setid)
            if changed or update_listing:
                collection.update_listings([dataset])
                log.verbose('%s listing rendered', setid)
        finally:
            if not keep:
                for tmpdir, tmpdir_fd in store.pending.values():
                    store.logdebug('Cleaning up %r', tmpdir)
                    shutil.rmtree(tmpdir)
                    fcntl.flock(tmpdir_fd, fcntl.LOCK_UN)
                    os.close(tmpdir_fd)
                store.pending.clear()

        return changed

    def scan(self, dry_run=None):
        for collection in self.collections.values():
            for scanroot in collection.scanroots:
                collection.scan(scanroot, dry_run)

    def comment(self, username, message, ids):
        now = int(utils.now() * 1000)
        comments = [Comment(dataset_id=id, author=username, time_added=now,
                            text=message)
                    for id in ids]
        db.session.add_all(comments)
        db.session.commit()

    def tag(self, setids, add=None, remove=None):
        assert setids
        assert add or remove, (add, remove)
        add = sorted(add or [])
        remove = remove or []
        addremove = set(add) | set(remove)
        assert len(addremove) == len(add) + len(remove), (add, remove)

        query = db.session.query(Dataset.collection, Dataset.id, Dataset.setid)\
                          .filter(Dataset.setid.in_(setids))\
                          .order_by(Dataset.collection)
        for collection, group in groupby(query, key=lambda x: x[0]):
            setidmap = {setid: id for _, id, setid in group}
            dataset_ids = setidmap.values()

            if add:
                stmt = Tag.__table__.insert().prefix_with('OR IGNORE')
                db.session.execute(stmt, [{'collection': collection,
                                           'value': x} for x in add])

            tags = {value: id for id, value in (
                db.session.query(Tag.id, Tag.value)
                .filter(Tag.collection == collection)
                .filter(Tag.value.in_(addremove)))}

            if add:
                stmt = dataset_tag.insert().prefix_with('OR IGNORE')
                db.session.execute(stmt,
                                   [{'dataset_id': x, 'tag_id': y} for x, y in
                                    product(dataset_ids, (tags[x] for x in add))])

            if remove:
                where = (dataset_tag.c.tag_id.in_(tags[x] for x in remove) &
                         dataset_tag.c.dataset_id.in_(dataset_ids))
                stmt = dataset_tag.delete().where(where)
                db.session.execute(stmt)

        db.session.commit()


def dump_database(dburi):
    """Dump database.

    The structure of the database is reflected and therefore also
    older databases, not matching the current version of marv, can be
    dumped.
    """

    engine = sqla.create_engine(dburi)
    meta = sqla.MetaData(engine)
    meta.reflect()
    con = engine.connect()
    tables = {k: v for k, v in meta.tables.items() if not k.startswith('listing_')}
    select = sqla.sql.select

    def rows2dcts(stmt):
        result = con.execute(stmt)
        keys = result.keys()
        return [{k: v for k, v in zip(keys, row)} for row in result]

    comments = {}
    comment_t = tables.pop('comment')
    stmt = select([comment_t]).order_by(comment_t.c.dataset_id, comment_t.c.id)
    for did, grp in groupby(rows2dcts(stmt), key=lambda x: x['dataset_id']):
        assert did not in comments
        comments[did] = lst = []
        for comment in grp:
            # Everything except these fields is included in the dump
            del comment['dataset_id']
            del comment['id']
            lst.append(comment)

    files = {}
    file_t = tables.pop('file')
    stmt = select([file_t]).order_by(file_t.c.dataset_id, file_t.c.idx)
    for did, grp in groupby(rows2dcts(stmt), key=lambda x: x['dataset_id']):
        assert did not in files
        files[did] = lst = []
        for file in grp:
            # Everything except these fields is included in the dump
            del file['dataset_id']
            del file['idx']
            lst.append(file)

    tags = {}
    dataset_tag_t = tables.pop('dataset_tag')
    tag_t = tables.pop('tag')
    stmt = select([dataset_tag_t, tag_t]).where(dataset_tag_t.c.tag_id == tag_t.c.id)\
                                         .order_by(dataset_tag_t.c.dataset_id, tag_t.c.value)
    for did, grp in groupby(rows2dcts(stmt), key=lambda x: x['dataset_id']):
        assert did not in tags
        tags[did] = lst = []
        for tag in grp:
            del tag['dataset_id']
            del tag['collection']
            del tag['id']
            del tag['tag_id']
            lst.append(tag.pop('value'))
            assert not tag

    dump = {}
    dump['datasets'] = collections = {}
    dataset_t = tables.pop('dataset')
    stmt = select([dataset_t]).order_by(dataset_t.c.setid)
    for dataset in rows2dcts(stmt):
        did = dataset.pop('id')  # Everything except this is included in the dump
        dataset['comments'] = comments.pop(did, [])
        dataset['files'] = files.pop(did)
        dataset['tags'] = tags.pop(did, [])
        collections.setdefault(dataset.pop('collection'), []).append(dataset)

    assert not comments, comments
    assert not files, files
    assert not tags, tags

    groups = {}
    group_t = tables.pop('group')
    user_group_t = tables.pop('user_group')
    stmt = select([user_group_t, group_t]).where(user_group_t.c.group_id == group_t.c.id)\
                                          .order_by(user_group_t.c.user_id, group_t.c.name)
    for uid, grp in groupby(rows2dcts(stmt), key=lambda x: x['user_id']):
        assert uid not in groups
        groups[uid] = lst = []
        for group in grp:
            del group['user_id']
            del group['group_id']
            del group['id']
            name = group.pop('name')
            assert not group
            lst.append(name)

    dump['users'] = users = []
    user_t = tables.pop('user')
    stmt = select([user_t]).order_by(user_t.c.name)
    for user in rows2dcts(select([user_t])):
        user_id = user.pop('id')  # Everything except this is included in the dump
        user['groups'] = groups.pop(user_id, [])
        users.append(user)

    assert not groups, groups
    assert not tables, tables.keys()

    return dump



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
import sys

import marv.app
import marv.site
from marv_cli import setup_logging
setup_logging(os.environ.get('MARV_LOGLEVEL', 'info'))

config = os.environ['MARV_CONFIG']
app_root = os.environ.get('MARV_APPLICATION_ROOT') or '/'
init = os.environ.get('MARV_INIT')

site = marv.site.Site(config)
site.load_for_web()
try:
    application = marv.app.create_app(site, app_root=app_root, init=init)
except OSError as e:
    if e.errno == 13:
        print(e, file=sys.stderr)
        sys.exit(13)
    raise



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest

from ..utils import decode_setid, encode_setid


class TestCase(unittest.TestCase):
    def test_encode_decode_setid(self):
        setid = 2**128 - 1
        encoded = encode_setid(setid)
        decoded = decode_setid(encoded.lower())
        self.assertEqual(encoded, '77777777777777777777777774')
        self.assertEqual(decoded, setid)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import flask

from marv.model import Dataset, db
from .tooling import api_endpoint as marv_api_endpoint


@marv_api_endpoint('/dataset', methods=['DELETE'])
def delete():
    datasets = Dataset.__table__
    ids = flask.request.get_json()
    if not ids:
        flask.abort(400)
    stmt = datasets.update()\
                   .where(datasets.c.id.in_(ids))\
                   .values(discarded=True)
    db.session.execute(stmt)
    db.session.commit()
    return flask.jsonify({})



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import unittest
from collections import defaultdict
from itertools import product

from ..testing import make_dataset, make_sink, marv

from ..run import run_nodes
from ..stream import Handle


dataset = make_dataset()
SETID = dataset.setid


class TestCase(unittest.TestCase):
    def test_plain_is_error(self):
        @marv.node()
        def source():
            yield 1

        sink = make_sink(source)
        with self.assertRaises(RuntimeError):  # TODO: propert exc
            run_nodes(dataset, [sink], {})

    def test_one_source(self):
        @marv.node()
        def source():
            yield marv.push(1)
            yield marv.push(2)
            yield marv.push(3)

        sink = make_sink(source)
        run_nodes(dataset, [sink], {})
        self.assertEqual(sink.stream, [1, 2, 3])

    def test_one_consumer(self):
        @marv.node()
        def source():
            yield marv.push(1)
            yield marv.push(2)
            yield marv.push(3)

        @marv.node()
        def cubic():
            stream = yield marv.get_stream(source)
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                yield marv.push(msg**3)

        sink = make_sink(cubic)
        run_nodes(dataset, [sink], {})
        self.assertEqual(sink.stream, [1, 8, 27])

    def test_two_consumers(self):
        @marv.node()
        def source():
            yield marv.push(1)
            yield marv.push(2)
            yield marv.push(3)

        @marv.node()
        def cubic():
            stream = yield marv.get_stream(source)
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg**3)

        source_sink = make_sink(source)
        cubic_sink = make_sink(cubic)
        run_nodes(dataset, [source_sink, cubic_sink], {})
        self.assertEqual(source_sink.stream, [1, 2, 3])
        self.assertEqual(cubic_sink.stream, [1, 8, 27])

    def test_ondemand_group(self):
        @marv.node(group='ondemand')
        def source():
            requested = yield marv.get_requested()
            self.assertIn(Handle(SETID, source, 'evensub'), requested)
            self.assertIn(Handle(SETID, source, 'oddsub'), requested)
            self.assertIn(Handle(SETID, source, 'primesub'), requested)

            out = {x.name: marv.create_stream(x.name) for x in requested}
            for k, stream in out.items():
                out[k] = yield stream
            self.assertEqual(sorted(out.keys()), ['evensub', 'oddsub', 'primesub'])

            for i in range(1, 6):
                yield marv.push(out['oddsub' if i % 2 else 'evensub'].msg(i))
                if i in [2, 3, 5]:
                    yield marv.push(out['primesub'].msg(i))

            # How do we output one message for multiple streams?

        @marv.node()
        def even():
            stream = yield marv.get_stream(source, 'evensub')
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        @marv.node()
        def odd():
            stream = yield marv.get_stream(source, 'oddsub')
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        @marv.node()
        def prime():
            stream = yield marv.get_stream(source, 'primesub')
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        even_sink = make_sink(even)
        odd_sink = make_sink(odd)
        prime_sink = make_sink(prime)
        run_nodes(dataset, [even_sink, odd_sink, prime_sink], {})
        self.assertEqual(even_sink.stream, [2, 4])
        self.assertEqual(odd_sink.stream, [1, 3, 5])
        self.assertEqual(prime_sink.stream, [2, 3, 5])

    def test_ondemand_group_with_restart(self):
        reqs = []
        @marv.node(group='ondemand')
        def source():
            requested = yield marv.get_requested()
            reqs.append(requested)
            creates = [marv.create_stream(x.name) for x in requested]
            streams = []
            for create in creates:
                stream = yield create
                streams.append(stream)
            msgs = list(product(streams, [1, 2]))
            for stream, msg in msgs:
                yield marv.push(stream.msg(msg))

        msgs = []
        @marv.node()
        def consumer():
            stream1 = yield marv.get_stream(source, 'a')
            stream2 = yield marv.get_stream(source, 'b')
            streams = [stream1, stream2]
            while streams:
                for stream in streams[:]:
                    msg = yield marv.pull(stream)
                    if msg is None:
                        streams.remove(stream)
                        continue
                    msgs.append((stream.name, msg))

        run_nodes(dataset, [consumer], {})
        self.assertEqual(reqs, [
            [Handle(SETID, source, 'a')],
            [Handle(SETID, source, 'a'), Handle(SETID, source, 'b')],
        ])
        self.assertEqual(msgs, [('a', 1), ('b', 1), ('a', 2), ('b', 2)])

    def test_ondemand_group_with_restart_and_foreach(self):
        reqs = []
        @marv.node(group='ondemand')
        def source():
            requested = yield marv.get_requested()
            reqs.append(requested)
            creates = [marv.create_stream(x.name) for x in requested]
            streams = []
            for create in creates:
                stream = yield create
                streams.append(stream)
            msgs = list(product(streams, [1,2]))
            for stream, msg in msgs:
                yield marv.push(stream.msg(msg))

        @marv.node()
        def streamA():
            stream = yield marv.get_stream(source, 'a')
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        @marv.node()
        def streamB():
            stream = yield marv.get_stream(source, 'b')
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        @marv.node(group=True)
        def merged():
            stream1 = yield marv.get_stream(streamA)
            stream2 = yield marv.get_stream(streamB)
            yield marv.push(stream1)
            yield marv.push(stream2)

        msgs = []
        @marv.node()
        @marv.input('stream', foreach=merged)
        def consumer(stream):
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msgs.append((stream, msg))

        run_nodes(dataset, [consumer], {})
        self.maxDiff = None
        self.assertEqual(reqs, [[Handle(SETID, source, 'a')],
                                [Handle(SETID, source, 'a'), Handle(SETID, source, 'b')]])
        self.assertEqual(msgs, [(Handle(SETID, streamA, 'default'), 1),
                                (Handle(SETID, streamA, 'default'), 2),
                                (Handle(SETID, streamB, 'default'), 1),
                                (Handle(SETID, streamB, 'default'), 2)])

    def test_request_input(self):
        started = defaultdict(bool)
        finished = defaultdict(bool)

        @marv.node(group=True)
        def multi():
            started[multi] = True
            a_out = yield marv.create_stream('a')
            b_out = yield marv.create_stream('b')
            self.assertEqual(a_out, Handle(SETID, multi, 'a'))
            self.assertEqual(b_out, Handle(SETID, multi, 'b'))

            yield a_out.msg(1)
            yield b_out.msg(10)
            yield a_out.msg(2)
            yield b_out.msg(20)
            finished[multi] = True

        @marv.node()
        def consumer():
            started[consumer] = True
            node = yield marv.get_stream(multi)
            a_in, b_in = yield marv.pull_all(node, node)
            self.assertEqual(a_in, Handle(SETID, multi, 'a'))
            self.assertEqual(b_in, Handle(SETID, multi, 'b'))
            # streams = yield marv.io(*node)  # like for values
            # streams = yield marv.io(node, consume=True)  # a bit less voodoo

            values = []
            while True:
                msgs = yield marv.pull_all(a_in, b_in)
                if msgs == [None, None]:
                    break
                values.append(msgs)
            self.assertEqual(values, [[1, 10], [2, 20]])
            finished[consumer] = True

        run_nodes(dataset, [consumer], {})
        self.assertTrue(started[multi])
        self.assertTrue(started[consumer])
        self.assertTrue(finished[multi])
        self.assertTrue(finished[consumer])

    def test_fork(self):
        @marv.node(group=True)
        def source():
            out1 = yield marv.create_stream('Output 1')
            out2 = yield marv.create_stream('Output 2')
            out3 = yield marv.create_stream('Output 3')
            yield out1.msg(a=1)
            yield out2.msg(b=10)
            yield out3.msg(c=100)
            yield out1.msg(a=2)
            yield out2.msg(b=20)
            yield out3.msg(c=200)

        foreach_msgs = defaultdict(list)

        @marv.node(group=True)
        def forking(stream=None):
            node = yield marv.get_stream(source)  # UX not nice, naming

            if stream is None:
                stream = yield marv.pull(node)
                other = yield marv.pull(node)
                while other is not None:
                    name = other.key[-1]
                    yield marv.fork(name, inputs={'stream': other}, group=False)
                    other = yield marv.pull(node)

            msgs = foreach_msgs[stream.key]
            msg = yield marv.pull(stream)
            while msg:
                msgs.append(msg)
                msg = yield marv.pull(stream)

        run_nodes(dataset, [forking], {})
        self.assertEqual(dict(foreach_msgs), {
            (SETID, source, 'Output 1'): [{'a': 1}, {'a': 2}],
            (SETID, source, 'Output 2'): [{'b': 10}, {'b': 20}],
            (SETID, source, 'Output 3'): [{'c': 100}, {'c': 200}],
        })

    def test_foreach(self):
        @marv.node(group=True)
        def source():
            out1 = yield marv.create_stream('Output 1', foo=1)
            out2 = yield marv.create_stream('Output 2', foo=2)
            out3 = yield marv.create_stream('Output 3', foo=3)
            yield out1.msg(a=1)
            yield out2.msg(b=10)
            yield out3.msg(c=100)
            yield out1.msg(a=2)
            yield out2.msg(b=20)
            yield out3.msg(c=200)

        foreach_msgs = defaultdict(list)
        @marv.node()
        @marv.input('stream', foreach=source)
        def foreach(stream):
            assert stream.key not in foreach_msgs
            msgs = foreach_msgs[stream.key]
            msg = yield marv.pull(stream)
            while msg:
                msgs.append(msg)
                msg = yield marv.pull(stream)

        streams = []
        all_msgs = []
        @marv.node()
        @marv.input('node', default=source)
        def withall(node):
            while True:
                stream = yield marv.pull(node)
                if stream is None:
                    break
                streams.append(stream)

            msgs = yield marv.pull_all(*streams)
            while filter(None, msgs):
                all_msgs.append(msgs)
                msgs = yield marv.pull_all(*streams)

        run_nodes(dataset, [foreach, withall], {})
        self.assertEqual(streams, [Handle(SETID, source, 'Output 1'),
                                   Handle(SETID, source, 'Output 2'),
                                   Handle(SETID, source, 'Output 3')])
        self.assertEqual(streams[0].header['foo'], 1)
        self.assertEqual(streams[1].header['foo'], 2)
        self.assertEqual(streams[2].header['foo'], 3)
        self.assertEqual(all_msgs, [[{'a': 1}, {'b': 10}, {'c': 100}],
                                    [{'a': 2}, {'b': 20}, {'c': 200}]])
        self.maxDiff = None
        self.assertEqual(dict(foreach_msgs), {
            (SETID, source, 'Output 1'): [{'a': 1}, {'a': 2}],
            (SETID, source, 'Output 2'): [{'b': 10}, {'b': 20}],
            (SETID, source, 'Output 3'): [{'c': 100}, {'c': 200}],
        })

    def test_foreach_with_header(self):
        @marv.node(group=True)
        def source():
            out1 = yield marv.create_stream('Output 1', foo=1)
            yield out1.msg(a=1)
            yield out1.msg(a=2)

        @marv.node()
        @marv.input('stream', foreach=source)
        def foreach(stream):
            yield marv.set_header(**stream.header)
            while True:
                msg = yield marv.pull(stream)
                if msg is None:
                    break
                msg = yield marv.push(msg)

        foreach_msgs = defaultdict(list)
        foreach_streams = []
        @marv.node()
        @marv.input('foreaches', default=foreach)
        def consumer(foreaches):
            while True:
                stream = yield marv.pull(foreaches)
                if stream is None:
                    break
                foreach_streams.append(stream)
                msgs = foreach_msgs[stream.key]
                while True:
                    msg = yield marv.pull(stream)
                    if msg is None:
                        break
                    msgs.append(msg)

        run_nodes(dataset, [consumer], {})
        self.assertEqual(len(foreach_streams), 1)
        handle = foreach_streams[0]
        self.assertEqual(handle, Handle(SETID, foreach, '0'))
        self.assertEqual(handle.foo, 1)
        self.maxDiff = None
        self.assertEqual(dict(foreach_msgs), {
            (SETID, foreach, '0'): [{'a': 1}, {'a': 2}],
        })

    def test_create_stream(self):
        @marv.node(group=True)
        def source():
            out1 = yield marv.create_stream('Output 1')
            out2 = yield marv.create_stream('Output 2', foo=1)

        handles = []
        @marv.node()
        @marv.input('source', source)
        def consumer(source):
            stream = yield marv.pull(source)
            handles.append(stream)
            stream = yield marv.pull(source)
            handles.append(stream)

        run_nodes(dataset, [consumer], {})
        self.assertEqual(handles[0].name, 'Output 1')
        self.assertEqual(handles[1].name, 'Output 2')
        self.assertEqual(handles[1].foo, 1)
        with self.assertRaises(AttributeError):
            handles[0].foo



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from itertools import cycle

import flask

from marv.model import Tag, dataset_tag, db
from .tooling import api_endpoint as marv_api_endpoint


@marv_api_endpoint('/tag', methods=['POST'])
def tag():
    # TODO: very similar to cli marv_tag
    changes = flask.request.get_json()
    if not changes:
        flask.abort(400)

    for collection, ops in changes.items():
        if collection not in flask.current_app.site.collections:
            flask.abort(400)

        addop = ops.get('add', {})
        removeop = ops.get('remove', {})

        add = addop.viewkeys()
        remove = removeop.viewkeys()

        if add:
            stmt = Tag.__table__.insert().prefix_with('OR IGNORE')
            db.session.execute(stmt, [{'collection': collection,
                                       'value': x} for x in add])

        if add or remove:
            tags = {value: id for id, value in (db.session.query(Tag.id, Tag.value)
                                                .filter(Tag.collection == collection)
                                                .filter(Tag.value.in_(add | remove)))}

        if add:
            stmt = dataset_tag.insert().prefix_with('OR IGNORE')
            values = [{'tag_id': x, 'dataset_id': y} for tag, ids in addop.items()
                      for x, y in zip(cycle([tags[tag]]), ids)]
            db.session.execute(stmt, values)

        if remove:
            where = reduce(lambda acc, x: acc | x, (
                ((dataset_tag.c.tag_id == x) & (dataset_tag.c.dataset_id == y))
                for tag, ids in removeop.items()
                for x, y in zip(cycle([tags[tag]]), ids)
            ))
            stmt = dataset_tag.delete().where(where)
            db.session.execute(stmt)

    db.session.commit()

    # TODO: report about unprocessed "setids"
    return flask.jsonify({})



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import json
import os
from collections import OrderedDict, deque

from marv_pycapnp import Wrapper
from marv_node.io import THEEND
from marv_node.stream import Handle, Msg, RequestedMessageTooOld, Stream
from marv_nodes.types_capnp import File


class ReadStream(Stream):
    streamfile = None

    def __init__(self, prehandle, streamdir, setdir, msgs=None, info=None):
        header = info['header'] if info else {}
        handle = Handle(*prehandle.key, group=prehandle.group, header=header)
        self.handle = handle
        self.cache = deque([Msg(idx=-1, handle=handle, data=handle)], self.CACHESIZE)
        if handle.group:
            self.stream = iter([
                Handle(handle.setid, handle.node, x['name'],
                       group=bool(x['streams']),
                       header=x['header'])
                for x in info['streams'].values()
            ])
        elif msgs:
            self.stream = (Wrapper(x._reader, streamdir, setdir) for x in msgs)
        else:
            assert not info['streams']
            path = os.path.join(streamdir, '{}-stream'.format(handle.name))
            streamfile = open(path, 'rb')
            self.streamfile = streamfile
            self.stream = (Wrapper(x, streamdir, setdir) for x in
                           handle.node.schema.read_multiple_packed(streamfile))

    def get_msg(self, req):
        self.logdebug('got %r', req)
        assert req.handle == self.handle

        latest_idx = self.cache[0].idx
        offset = latest_idx - req.idx if self.cache else -1
        if offset < 0:
            assert offset == -1  # for now
            try:
                data = self.stream.next()
            except StopIteration:
                data = THEEND
                if self.streamfile:
                    self.streamfile.close()
                self.ended = True
            msg = Msg(latest_idx + 1, req.handle, data)
            self.cache.appendleft(msg)
        else:
            try:
                msg = self.cache[offset]
            except IndexError:
                raise RequestedMessageTooOld(req, offset)
        assert msg.data is not None
        self.logdebug('return %r', msg)
        return msg


class PersistentStream(Stream):
    ended = False

    def __init__(self, handle, streamdir, setdir, commit, parent=None):
        self.parent = parent
        self.handle = handle
        self.streamdir = streamdir
        self.setdir = setdir
        self.cache = deque((), self.CACHESIZE)
        self.streams = OrderedDict() if self.group else None
        # open right away: empty stream -> empty file
        path = os.path.join(streamdir, '{}-stream'.format(handle.name))
        self.streamfile = None if self.group else open(path, 'wb')
        self.done = set()
        self._commit = commit

    def add_msg(self, msg):
        assert not self.ended
        assert msg.handle == self.handle
        assert msg.idx is not None
        if self.group:
            assert isinstance(msg.data, Handle) or msg.data is THEEND, (self, msg)
        else:
            assert not isinstance(msg.data, Handle) or msg.idx == -1, (self, msg)

        expected_idx = self.cache[0].idx + 1 if self.cache else -1
        assert msg.idx == expected_idx, (msg.idx, expected_idx)

        self.logdebug('adding %r', msg)
        if msg.data is THEEND:
            self.ended = True
            self.logdebug('ended')
            if not self.group or self.done == self.streams.viewkeys():
                self._commit(self)
        elif isinstance(msg.data, Wrapper) and \
             msg.data._reader.schema.node.id == File.schema.node.id:
            assert msg.data._streamdir == self.streamdir
            stat = os.stat(msg.data.path)
            mtime = int(stat.st_mtime * 1e9)
            msg.data._reader = File.new_message(path=msg.data._reader.path,
                                                mtime=mtime, size=stat.st_size)\
                                   .as_reader()
        self.cache.appendleft(msg)
        if not self.group and msg.idx >= 0:
            # TODO: Maybe we want to have a builder and change to
            # reader before putting in cache?
            try:
                write_packed = msg.data._reader.as_builder().write_packed
            except AttributeError:
                assert msg.data is THEEND
                self.streamfile.close()
                return
            write_packed(self.streamfile)

    def get_msg(self, req):
        assert req.handle == self.handle

        offset = self.cache[0].idx - req.idx if self.cache else -1
        if offset < 0:
            assert not self.ended
            return None

        try:
            msg = self.cache[offset]
        except IndexError:
            raise RequestedMessageTooOld(req, offset)
        assert msg.data is not None
        self.logdebug('return %r', msg)
        return msg

    def create_stream(self, name, group, header=None):
        assert self.group
        assert not self.ended
        handle = Handle(self.handle.setid, self.handle.node, name, group=group, header=header)
        stream = type(self)(handle, self.streamdir, self.setdir,
                            self.commit_substream, parent=self)
        self.streams[handle.name] = stream
        return stream

    def commit_substream(self, substream):
        self.done.add(substream.handle.name)
        if self.ended and self.done == self.streams.viewkeys():
            self._commit(self)

    def make_file(self, name):
        assert not self.ended
        assert not self.group, (self, name)
        path = os.path.join(self.streamdir, name)
        fd = os.open(path, os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o666)
        os.close(fd)  # shall we pass on the open handle?
        # TODO: consider storing path relative to setdir instead of name
        file = File.new_message(path=path).as_reader()
        wrapper = Wrapper(file, self.streamdir, self.setdir)
        self.logdebug('made file %r', wrapper)
        return wrapper



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import OrderedDict

import flask
from flask import current_app, request


class APIEndpoint(object):
    MIN_API_VERSION = 3
    LATEST_API_VERSION = 3
    HEADER_PREFIX = "application/vnd.marv.v"
    acl = None

    def __init__(self, name, func, url_rule, defaults=None, methods=None, version=None, force_acl=None):
        self.name = name
        version = self.MIN_API_VERSION if version is None else version
        self.funcs = [(version, func)]
        self.url_rules = [(url_rule, {'defaults': defaults, 'methods': methods})]
        self._force_acl = set(force_acl) if force_acl else None

    def __call__(self, *args, **kw):
        authorization = request.headers.get('Authorization')
        # TODO: can authorization be '' or is None test?
        if not authorization:
            authorization = request.args.get('access_token')
        groups = current_app.um.check_authorization(self.acl, authorization)

        try:
            accepted = (x[0] for x in flask.request.accept_mimetypes
                        if x[0].startswith(self.HEADER_PREFIX)).next()
            accepted_version = int(accepted[len(self.HEADER_PREFIX):])
        except (StopIteration, ValueError):
            accepted_version = self.MIN_API_VERSION

        try:
            func = (func for version, func in self.funcs
                    if version <= accepted_version).next()
        except StopIteration:
            flask.abort(406)

        return func(*args, **kw)

    def init_app(self, app, url_prefix=None, name_prefix=None):
        self.acl = self._force_acl if self._force_acl else set(current_app.acl[self.name])
        name = '.'.join(filter(None, [name_prefix, self.name]))
        for url_rule, options in self.url_rules:
            url_rule = ''.join(filter(None, [url_prefix, url_rule]))
            app.add_url_rule(url_rule, name, self, **options)


class APIGroup(object):
    def __init__(self, name, func, url_prefix=None):
        self.name = name
        self.func = func
        self.url_prefix = url_prefix
        self.endpoints = OrderedDict()

    def add_endpoint(self, ep):
        """endpoints and groups are all the same (for now)"""
        assert ep.name not in self.endpoints, ep
        self.endpoints[ep.name] = ep

    def endpoint(self, *args, **kw):
        return api_endpoint(*args, registry=self.endpoints, **kw)

    def init_app(self, app, url_prefix=None, name_prefix=None):
        self.func(app)
        name_prefix = '.'.join(filter(None, [name_prefix, self.name]))
        url_prefix = '/'.join(filter(None, [url_prefix, self.url_prefix])) or None
        for ep in self.endpoints.values():
            ep.init_app(app, url_prefix=url_prefix, name_prefix=name_prefix)

    def __repr__(self):
        return '<APIGroup {} url_prefix={}>'.format(self.name, self.url_prefix)


def api_endpoint(url_rule, defaults=None, methods=None, version=None,
                 cls=APIEndpoint, registry=None, force_acl=None):
    def decorator(func):
        if isinstance(func, cls):
            func.url_rules.append((url_rule, {'defaults': defaults, 'methods': methods}))
            return func

        name = func.func_name
        rv = cls(name, func, url_rule=url_rule, defaults=defaults,
                 methods=methods, version=version, force_acl=force_acl)
        rv.__doc__ = func.__doc__

        if registry is not None:
            assert name not in registry, name
            registry[name] = rv
        return rv
    return decorator


def api_group(url_prefix=None, cls=APIGroup):
    def decorator(func):
        if isinstance(func, cls):
            raise TypeError('Attempted to convert function into api group twice.')

        name = func.func_name
        rv = cls(name, func, url_prefix)
        rv.__doc__ = func.__doc__
        return rv
    return decorator



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from collections import defaultdict
from itertools import product

from ..testing import make_dataset, make_sink, marv

from ..run import run_nodes
from ..stream import Handle


dataset = make_dataset()
SETID = dataset.setid


@marv.node()
def meta():
    """Meta information about a dataset"""
    yield marv.push({'topics': ['b', 'c']})


@marv.node(group='ondemand')
@marv.input('meta', default=meta)
def messages(meta):
    """Produce streams for requested topics if they exist"""
    meta = yield marv.pull(meta)
    requested = yield marv.get_requested()
    all_msgs = {'a': range(1),
                'b': range(2),
                'c': range(3)}
    topics = [x.name for x in requested]
    streams = {}
    for topic in topics:
        stream = yield marv.create_stream(topic)
        if topic in meta['topics']:
            streams[topic] = stream
        else:
            yield stream.finish()

    while streams:
        for topic, stream in streams.items():
            msgs = all_msgs[topic]
            msg = msgs.pop(0)
            yield stream.msg('{}{}'.format(topic, msg))
            if not msgs:
                del streams[topic]


@marv.node()
@marv.input('a', default=marv.select(messages, 'a'))
def nodeA(a):
    yield marv.set_header(topic='a')
    while True:
        msg = yield marv.pull(a)
        if msg is None:
            return
        yield marv.push('nodeA-{}'.format(msg))


@marv.node()
@marv.input('b', default=marv.select(messages, 'b'))
def nodeB(b):
    yield marv.set_header(topic='b')
    while True:
        msg = yield marv.pull(b)
        if msg is None:
            return
        yield marv.push('nodeB-{}'.format(msg))


@marv.node()
@marv.input('c', default=marv.select(messages, 'c'))
def nodeC(c):
    yield marv.set_header(topic='c')
    while True:
        msg = yield marv.pull(c)
        if msg is None:
            return
        yield marv.push('nodeC-{}'.format(msg))


@marv.node()
@marv.input('nodeA', default=nodeA)
@marv.input('nodeB', default=nodeB)
@marv.input('nodeC', default=nodeC)
def collect(nodeA, nodeB, nodeC):
    yield marv.set_header()
    acc = []
    streams = [nodeA, nodeB, nodeC]
    while streams:
        for stream in streams[:]:
            msg = yield marv.pull(stream)
            if msg is None:
                streams.remove(stream)
                continue
            acc.append(msg)
    yield marv.push({'acc': acc})


def test():
    sink = make_sink(collect)
    run_nodes(dataset, [sink], {})
    assert sink.stream == [{'acc': ['nodeB-b0', 'nodeC-c0', 'nodeB-b1', 'nodeC-c1', 'nodeC-c2']}]



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

from pkg_resources import iter_entry_points

from .auth import auth
from .comment import comment
from .dataset import dataset
from .delete import delete
from .tag import tag
from .collection import collection, meta
from .tooling import api_group as marv_api_group


@marv_api_group()
def webapi(app):
    pass


# Groups and endpoints are all the same for now
webapi.add_endpoint(auth)
webapi.add_endpoint(comment)
webapi.add_endpoint(dataset)
webapi.add_endpoint(delete)
webapi.add_endpoint(collection)
webapi.add_endpoint(meta)
webapi.add_endpoint(tag)


from marv_robotics.webapi import robotics
webapi.add_endpoint(robotics)


for ep in iter_entry_points(group='marv_webapi'):
    endpoint = ep.load()
    webapi.add_endpoint(endpoint)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import os
import sys

from capnp.lib.capnp import KjException

import marv
from marv_detail import Widget
from marv_node.setid import SetID
from marv_pycapnp import Wrapper
from .types_capnp import Dataset


@marv.node(Dataset)
def dataset():
    raise RuntimeError('Datasets are added not run')
    yield


def load_dataset(setdir, dataset):
    setid = SetID(dataset.setid)
    files = [{'path': x.path,
              'missing': bool(x.missing),
              'mtime': x.mtime * 10**6,
              'size': x.size} for x in sorted(dataset.files, key=lambda x: x.idx)]
    dct = {'id': setid,
           'name': dataset.name,
           'files': files,
           'time_added': dataset.time_added * 10**6,
           'timestamp': dataset.timestamp * 10**6}
    try:
        wrapper = Wrapper.from_dict(Dataset, dct)
    except KjException as e:
        from pprint import pformat
        print('Schema violation for %s with data:\n%s\nschema: %s' % (
            Dataset.schema.node.displayName,
            pformat(dct),
            Dataset.schema.node.displayName), file=sys.stderr)
        raise e
    wrapper._setdir = setdir  # needed by dataset.load(node)
    return [wrapper]
dataset.load = load_dataset


@marv.node(Widget)
@marv.input('dataset', default=dataset)
def summary_keyval(dataset):
    dataset = yield marv.pull(dataset)
    if len(dataset.files) < 2:
        return
    yield marv.push({'keyval': {
        'items': [
            {'title': 'size', 'formatter': 'filesize', 'list': False,
             'cell': {'uint64': sum(x.size for x in dataset.files)}},
            {'title': 'files', 'list': False,
             'cell': {'uint64': len(dataset.files)}},
        ]
    }})


@marv.node(Widget)
@marv.input('dataset', default=dataset)
def meta_table(dataset):
    dataset = yield marv.pull(dataset)
    columns = [
        {'title': 'Name', 'formatter': 'rellink'},
        {'title': 'Size', 'formatter': 'filesize'},
    ]
    # dataset.id is setid here
    rows = [{'id': idx, 'cells': [
        {'link': {'href': '{}'.format(idx),
                  'title': os.path.basename(f.path)}},
        {'uint64': f.size},
    ]} for idx, f in enumerate(dataset.files)]
    yield marv.push({'table': {'columns': columns, 'rows': rows}})



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import functools
import hashlib
import os
from base64 import b32encode
from collections import OrderedDict, namedtuple
from inspect import isgeneratorfunction
from itertools import count, product

from .io import fork, get_logger, get_stream, pull
from .mixins import Keyed


class InputNameCollision(Exception):
    pass


def input(name, default=None, foreach=None):
    """Decorator to declare input for a node.

    Plain inputs, that is plain python objects, are directly passed to
    the node. Whereas streams generated by other nodes are requested
    and once the handles of all input streams are available the node
    is instantiated.

    Args:
        name (str): Name of the node function argument the input will
            be passed to.
        default: An optional default value for the input. This can be
            any python object or another node.
        foreach (bool): This parameter is currently not supported and
            only for internal usage.

    Returns:
        The original function decorated with this input
        specification. A function is turned into a node by the
        :func:`node` decorator.

    """
    assert default is None or foreach is None
    value = foreach if foreach is not None else default
    value = StreamSpec(value) if isinstance(value, Node) else value
    foreach = foreach is not None
    spec = InputSpec(name, value, foreach)
    def deco(func):
        """Add {!r} to function.""".format(spec)
        specs = func.__dict__.setdefault('__marv_input_specs__', OrderedDict())
        if spec.name in specs:
            raise InputNameCollision(spec.name)
        specs[spec.name] = spec
        return func
    return deco


def node(schema=None, header=None, group=None, version=None):
    """Turn function into node.

    Args:
        schema: capnproto schema describing the output messages format
        header: This parameter is currently not supported and only for
            internal usage.
        group (bool): A boolean indicating whether the default stream
            of the node is a group, meaning it will be used to
            published handles for streams or further groups. In case
            of :paramref:`marv.input.foreach` specifications this flag will
            default to `True`. This parameter is currently only for
            internal usage.
        version (int): This parameter currently has no effect.

    Returns:
        A :class:`Node` instance according to the given
        arguments and :func:`input` decorators.
    """
    def deco(func):
        """Turn function into node with given arguments.

        :func:`node`(schema={!r}, header={!r}, group={!r})
        """.format(schema, header, group)

        if isinstance(func, Node):
            raise TypeError('Attempted to convert function into node twice.')
        assert isgeneratorfunction(func), \
            "Node {}:{} needs to be a generator function".format(func.__module__,
                                                                 func.__name__)

        specs = getattr(func, '__marv_input_specs__', None)
        if hasattr(func, '__marv_input_specs__'):
            del func.__marv_input_specs__

        node = Node(func, schema=schema, header_schema=header,
                    group=group, specs=specs, version=version)
        functools.update_wrapper(node, func)
        return node
    return deco


_InputSpec = namedtuple('_InputSpec', ('name', 'value', 'foreach'))
class InputSpec(Keyed, _InputSpec):
    @property
    def key(self):
        value = self.value.key if hasattr(self.value, 'key') else self.value
        return (self.name, value, self.foreach)

    def clone(self, value):
        cls = type(self)
        value = StreamSpec(value) if isinstance(value, Node) else value
        return cls(self.name, value, self.foreach)

    def __repr__(self):
        foreach = 'foreach ' if self.foreach else ''
        return ('<{} {}{}={!r}>'.format(type(self), foreach, self.name, self.value))


class StreamSpec(object):
    def __init__(self, node, name=None):
        assert isinstance(node, Node)
        self.node = node
        self.name = name
        self.key = (node.key,) if name is None else (node.key, name)
        self.args = (node,) if name is None else (node, name)


class Node(Keyed):
    _key = None
    group = None

    @property
    def key(self):
        key = self._key
        if key:
            return key
        return '-'.join([self.specs_hash, self.fullname])

    @property
    def abbrev(self):
        return '.'.join([self.name, self.specs_hash[:10]])

    @staticmethod
    def genhash(specs):
        spec_keys = tuple(x.key for x in sorted(specs.values()))
        return b32encode(hashlib.sha256(repr(spec_keys)).digest()).lower()[:-4]

    def __init__(self, func, schema=None, header_schema=None, version=None,
                 name=None, namespace=None, specs=None, group=None):
        # TODO: assert no default values on func, or consider default
        # values (instead of) marv.input() declarations
        # Definitely not for node inputs as they are not passed directly!
        self.func = func
        self.version = version
        self.name = name = func.__name__ if name is None else name
        self.namespace = namespace = func.__module__ if namespace is None else namespace
        self.fullname = (name if not namespace else ':'.join([namespace, name]))
        self.specs_hash = self.genhash(specs or {})
        self.header_schema = header_schema
        self.schema = schema
        self.specs = specs or {}
        assert group in (None, False, True, 'ondemand'), group
        self.group = group if group is not None else \
                     any(x.foreach for x in self.specs.values())
        # TODO: StreamSpec, seriously?
        self.deps = {x.value.node for x in self.specs.values()
                     if isinstance(x.value, StreamSpec)}
        self.alldeps = self.deps.copy()
        self.alldeps.update(x for dep in self.deps
                            for x in dep.alldeps)

        self.consumers = set()
        for dep in self.deps:
            assert self not in dep.consumers, (dep, self)
            dep.consumers.add(self)

        self.dependent = set()
        for dep in self.alldeps:
            assert self not in dep.dependent, (dep, self)
            dep.dependent.add(self)

    def __call__(self, **inputs):
        return self.func(**inputs)

    def invoke(self, inputs=None):
        # We must not write any instance variables, a node is running
        # multiple times in parallel.

        common = []
        foreach_plain = []
        foreach_stream = []
        if inputs is None:
            for spec in self.specs.values():
                assert not isinstance(spec.value, Node), (self, spec.value)
                if isinstance(spec.value, StreamSpec):
                    value = yield get_stream(*spec.value.args)
                    target = foreach_stream if spec.foreach else common
                else:
                    value = spec.value
                    target = foreach_plain if spec.foreach else common
                target.append((spec.name, value))

        if foreach_plain or foreach_stream:
            log = yield get_logger()
            cross = product(*[[(k, x) for x in v] for k, v in foreach_plain])
            if foreach_stream:
                assert len(foreach_stream) == 1, self  # FOR NOW
                cross = list(cross)
                name, stream = foreach_stream[0]
                idx = count()
                while True:
                    value = yield pull(stream)
                    if value is None:
                        log.noisy('finished forking')
                        break
                    for inputs in cross:
                        inputs = dict(inputs)
                        inputs.update(common)
                        inputs[name] = value
                        # TODO: give fork a name
                        i = idx.next()
                        log.noisy('FORK %d with: %r', i, inputs)
                        yield fork('{}'.format(i), inputs, False)
            else:
                for i, inputs in enumerate(cross):
                    # TODO: consider deepcopy
                    inputs = dict(inputs)
                    inputs.update(common)
                    log.noisy('FORK %d with: %r', i, inputs)
                    yield fork('{}'.format(i), inputs, False)
                log.noisy('finished forking')
        else:
            if inputs is None:
                inputs = dict(common)
            while True:
                gen = self.func(**inputs)
                assert hasattr(gen, 'send')
                send = None
                while True:
                    send = yield gen.send(send)

    def clone(self, **kw):
        specs = {spec.name: (spec if spec.name not in kw else
                             spec.clone(kw.pop(spec.name)))
                 for spec in self.specs.values()}
        assert not kw, (kw, self.specs)
        cls = type(self)
        clone = cls(func=self.func, header_schema=self.header_schema,
                    schema=self.schema, specs=specs)
        return clone

    def __getitem__(self, key):
        from .tools import select
        import warnings
        warnings.warn('Use ``marv.select(node, name)`` instead of ``node[name]``',
                      DeprecationWarning, stacklevel=2)
        return select(self, key)

    def __str__(self):
        return self.key

    def __repr__(self):
        return '<Node {}>'.format(self.abbrev)



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

"""
Datasets are created based on information provided by scanners.  A
scanner is responsible to group files into named datasets::

    def scan(dirpath, dirnames, filenames):
        return [DatasetInfo(os.path.basename(x), [x])
                for x in filenames
                if x.endswith('.csv')]

Scanners are called for every directory within the configured
scanroots, while files and directories starting with a ``.`` and
directories containing an (empty) ``.marvignore`` file are ignored and
will not be traversed into.

Further, traversal into subdirectories can be controlled by
altering the :paramref:`.dirnames` list in-place. To block further
traversal, e.g. for a directory-based dataset type, set it to an
empty list -- :py:func:`os.walk` is used behind the scenes::

  dirnames[:] = []

"""

from __future__ import absolute_import, division, print_function

from collections import namedtuple

DatasetInfo = namedtuple('DatasetInfo', ('name', 'files'))



# -*- coding: utf-8 -*-
#
# Copyright 2016 - 2018  Ternaris.
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import absolute_import, division, print_function

import capnp

from marv_detail.types_capnp import Section, Widget
from marv_nodes.types_capnp import Dataset, File, GeoJson, Words
from marv_pycapnp.types_capnp import BoolValue, TextValue, DataValue
from marv_pycapnp.types_capnp import Float32Value, Float64Value
from marv_pycapnp.types_capnp import Int8Value, Int16Value, Int32Value, Int64Value
from marv_pycapnp.types_capnp import UInt8Value, UInt16Value, UInt32Value, UInt64Value
from marv_pycapnp.types_capnp import TimedBool, TimedText, TimedData
from marv_pycapnp.types_capnp import TimedFloat32, TimedFloat64
from marv_pycapnp.types_capnp import TimedInt8, TimedInt16, TimedInt32, TimedInt64
from marv_pycapnp.types_capnp import TimedUInt8, TimedUInt16, TimedUInt32, TimedUInt64

__all__ = (
    'Section', 'Widget',
    'Dataset', 'File', 'GeoJson', 'Words',
    'BoolValue', 'TextValue', 'DataValue',
    'Float32Value', 'Float64Value',
    'Int8Value', 'Int16Value', 'Int32Value', 'Int64Value',
    'UInt8Value', 'UInt16Value', 'UInt32Value', 'UInt64Value',
    'TimedBool', 'TimedText', 'TimedData',
    'TimedFloat32', 'TimedFloat64',
    'TimedInt8', 'TimedInt16', 'TimedInt32', 'TimedInt64',
    'TimedUInt8', 'TimedUInt16', 'TimedUInt32', 'TimedUInt64',
)

